"use strict";
var __awaiter = (this && this.__awaiter) || function (thisArg, _arguments, P, generator) {
    return new (P || (P = Promise))(function (resolve, reject) {
        function fulfilled(value) { try { step(generator.next(value)); } catch (e) { reject(e); } }
        function rejected(value) { try { step(generator["throw"](value)); } catch (e) { reject(e); } }
        function step(result) { result.done ? resolve(result.value) : new P(function (resolve) { resolve(result.value); }).then(fulfilled, rejected); }
        step((generator = generator.apply(thisArg, _arguments || [])).next());
    });
};
Object.defineProperty(exports, "__esModule", { value: true });
const assert = require("assert");
const request = require("request");
function setTimeoutPromise(time) {
    return new Promise((resolve) => setTimeout(() => resolve(), time));
}
function callServer(path, method, token) {
    return new Promise((resolve, reject) => {
        const reqOptions = {
            url: `https://api.github.com${path}`,
            method,
            headers: {
                'Accept': 'application/vnd.github.mister-fantastic-preview+json',
                'Authorization': token,
                'User-Agent': 'Unifile',
                'X-OAuth-Scopes': 'delete_repo, repo, user',
            },
        };
        request(reqOptions, (err, res, body) => {
            try {
                if (err) {
                    console.error('Github pages error', err);
                    reject(err);
                }
                else {
                    const result = JSON.parse(body);
                    resolve(result);
                }
            }
            catch (e) {
                console.error('Github pages error (try/catch)', err);
                reject(e);
            }
        });
    });
}
function HostingGhPages(unifile) {
    this.unifile = unifile;
    assert(this.unifile.listConnectors().find((connectorName) => connectorName === 'github'), 'Error: the Github service is required in order to activate the Github Pages hosting provider. You need to enable Github in unifile config, or disable Github Pages hosting provider (env var ENABLE_GITHUB_PAGES)');
}
exports.default = HostingGhPages;
HostingGhPages.prototype.getDefaultPageFileName = () => 'index.html';
HostingGhPages.prototype.finalizePublication = function (from, to, session, onStatus) {
    return setTimeoutPromise(2000)
        .then(() => {
        return new Promise((resolve, reject) => {
            try {
                const repo = to.path.split('/')[0];
                const owner = session.github.account.login;
                const path = `/repos/${owner}/${repo}/pages`;
                resolve(callServer(path, 'GET', session.github.token)
                    .then((result) => {
                    if (result.status) {
                        switch (result.status) {
                            case 'queued':
                                onStatus('Waiting for Github Pages to start deployment');
                                return this.finalizePublication(from, to, session, onStatus);
                                break;
                            case 'building':
                                onStatus('Deploying to Github Pages');
                                return this.finalizePublication(from, to, session, onStatus);
                                break;
                            case 'built':
                                onStatus('Done, the site is live on Github Pages');
                                return result.html_url;
                                break;
                            case 'errored':
                                onStatus('Github page build error');
                                throw new Error('Github page build error');
                                break;
                        }
                    }
                    else {
                        console.error('Unknown Github pages error', result);
                        reject(new Error(result.message || 'Unknown Github Pages error.'));
                    }
                }));
            }
            catch (e) {
                reject(e);
            }
        });
    });
};
HostingGhPages.prototype.getOptions = function (session) {
    const infos = this.unifile.getInfos(session, 'github');
    return {
        name: 'ghpages',
        displayName: 'Github Pages',
        isLoggedIn: infos.isLoggedIn,
        username: infos.username,
        authorizeUrl: '/ce/github/authorize',
        dashboardUrl: 'https://www.github.com',
        pleaseCreateAVhost: 'create an empty repository.',
        vhostsUrl: '/hosting/ghpages/vhost',
        buyDomainUrl: 'https://www.gandi.net',
        skipFolderSelection: true,
        skipVhostSelection: false,
        afterPublishMessage: 'Your website is now live.',
    };
};
HostingGhPages.prototype.getVhosts = function (session) {
    return __awaiter(this, void 0, void 0, function* () {
        const repos = yield (this.unifile.readdir(session, 'github', '/'));
        return repos
            .sort((a, b) => {
            return new Date(b.modified) - new Date(a.modified);
        })
            .map((file) => {
            return {
                name: file.name,
                domainUrl: `/hosting/ghpages/vhost/${file.name}`,
                skipDomainSelection: false,
                publicationPath: {
                    name: 'gh-pages',
                    folder: file.name,
                    path: `${file.name}/gh-pages`,
                    service: 'github',
                    url: `https://${session.github.account.login}.github.io/${file.name}/`,
                },
            };
        });
    });
};
HostingGhPages.prototype.getVhostData = (session, vhostName) => __awaiter(this, void 0, void 0, function* () {
    const owner = session.github.account.login;
    const path = `/repos/${owner}/${vhostName}/pages`;
    return callServer(path, 'GET', session.github.token)
        .then((result) => {
        return {
            domain: result.cname,
            url: result.html_url,
            status: result.status,
        };
    });
});
HostingGhPages.prototype.setVhostData = function (session, vhostName, data) {
    return __awaiter(this, void 0, void 0, function* () {
        if (data && data.domain && data.domain !== '') {
            return this.unifile.writeFile(session, 'github', `/${vhostName}/gh-pages/CNAME`, data.domain)
                .then(() => setTimeoutPromise(5000))
                .then(() => this.getVhostData(session, vhostName));
        }
        else {
            return this.unifile.unlink(session, 'github', `/${vhostName}/gh-pages/CNAME`)
                .catch((err) => {
                if (err.code !== 'ENOENT') {
                    console.error('Github pages error', err);
                    return Promise.reject(err);
                }
                return Promise.resolve();
            })
                .then(() => setTimeoutPromise(5000))
                .then(() => this.getVhostData(session, vhostName));
        }
    });
};
