"use strict";
Object.defineProperty(exports, "__esModule", { value: true });
const fs = require("fs");
const Path = require("path");
const Constants_1 = require("../../Constants");
const PACKAGE_JSON_DATA = JSON.parse(fs.readFileSync(Path.resolve(__dirname, '../../../../package.json')).toString());
const FRONT_END_VERSION = PACKAGE_JSON_DATA['version:frontend'].split('.').map((s) => parseInt(s));
const LATEST_VERSION = PACKAGE_JSON_DATA['version:backwardcompat'].split('.').map((s) => parseInt(s));
console.log(`\nSilex starts with backward compat version ${LATEST_VERSION} and front end version ${FRONT_END_VERSION}\n`);
class BackwardCompat {
    constructor(rootUrl) {
        this.rootUrl = rootUrl;
    }
    update(doc) {
        const styleTag = doc.querySelector('.' + Constants_1.Constants.JSON_STYLE_TAG_CLASS_NAME);
        if (styleTag) {
            styleTag.type = 'text/json';
        }
        let metaNode = doc.querySelector('meta[name="generator"]');
        if (!metaNode) {
            metaNode = doc.createElement('meta');
            metaNode.setAttribute('name', 'generator');
            doc.head.appendChild(metaNode);
        }
        const version = (metaNode.getAttribute('content') || '')
            .replace('Silex v', '')
            .split('.')
            .map((str) => parseInt(str, 10) || 0);
        const hasToUpdate = this.hasToUpdate(version, LATEST_VERSION);
        if (this.amIObsolete(version, LATEST_VERSION)) {
            return Promise.resolve('This website has been saved with a newer version of Silex. Continue at your own risks.');
        }
        else if (this.hasToUpdate(version, [2, 2, 7])) {
            return Promise.reject({
                message: 'This website has been saved with an older version of Silex, which is not supported anymore as of March 2018. In order to convert it to a newer version, please go to <a href="https://old.silex.me">old.silex.me</a> to open and then save your website. <a href="https://github.com/silexlabs/Silex/wiki/Website-saved-with-older-version-of-Silex">More about this here</a>',
            });
        }
        else if (hasToUpdate) {
            const allActions = {};
            return this.to2_2_8(version, doc).then((actions8) => {
                if (actions8.length) {
                    allActions['2.2.8'] = actions8;
                }
                return this.to2_2_9(version, doc).then((actions9) => {
                    if (actions9.length) {
                        allActions['2.2.9'] = actions9;
                    }
                    return this.to2_2_10(version, doc).then((actions10) => {
                        if (actions10.length) {
                            allActions['2.2.10'] = actions10;
                        }
                        this.updateStatic(doc);
                        metaNode.setAttribute('content', 'Silex v' + LATEST_VERSION.join('.'));
                        const report = Object.keys(allActions).map((_version) => {
                            return `<small>Update to version ${_version}:
              <ul>${allActions[_version].map((_action) => `<li class="no-list">${_action}</li>`).join('')}</ul>
          </small>`;
                        }).join('');
                        return `<h2>Website updated</h2>
          <p>This website has been updated to Silex latest version.</p>
          <p>Before you save it, please check that everything is fine. Saving it with another name could be a good idea too (menu file > save as).</p>
          <small>Bellow you will find details of what I did.</small>
          ${report}
        `;
                    });
                });
            });
        }
        else {
            this.updateStatic(doc);
            return Promise.resolve('');
        }
    }
    updateStatic(doc) {
        const elements = doc.querySelectorAll('[data-silex-static]');
        for (const element of elements) {
            const propName = element.src ? 'src' : 'href';
            const newUrl = this.getStaticResourceUrl(element[propName]);
            const oldUrl = element.getAttribute(propName);
            if (oldUrl !== newUrl) {
                element.setAttribute(propName, newUrl);
            }
        }
    }
    getStaticResourceUrl(url) {
        const pathRelativeToStaticMatch = url.match(/static\/[0-9]*\.[0-9]*\/(.*)/);
        if (pathRelativeToStaticMatch == null) {
            console.warn('Error: could not extract the path and file name of static asset', url);
            return url;
        }
        const pathRelativeToStatic = pathRelativeToStaticMatch[1];
        return `${this.rootUrl}/static/${FRONT_END_VERSION[0]}.${FRONT_END_VERSION[1]}/${pathRelativeToStatic}`;
    }
    amIObsolete(initialVersion, targetVersion) {
        return !!initialVersion[2] && initialVersion[0] > targetVersion[0] ||
            initialVersion[1] > targetVersion[1] ||
            initialVersion[2] > targetVersion[2];
    }
    hasToUpdate(initialVersion, targetVersion) {
        return initialVersion[0] < targetVersion[0] ||
            initialVersion[1] < targetVersion[1] ||
            initialVersion[2] < targetVersion[2];
    }
    to2_2_8(version, doc) {
        return new Promise((resolve, reject) => {
            const actions = [];
            if (this.hasToUpdate(version, [2, 2, 8])) {
                const menuButton = doc.querySelector('.menu-button');
                if (menuButton) {
                    menuButton.classList.remove('paged-element', 'paged-element-hidden', 'page-page-1', 'prevent-resizable');
                    menuButton.classList.add('hide-on-desktop');
                }
                doc.querySelector('.silex-inline-styles').innerHTML += '.silex-id-hamburger-menu {width: 50px;min-height: 40px;}';
                Array.from(doc.querySelectorAll('.page-element'))
                    .forEach((el) => {
                    el.setAttribute('href', '#!' + el.getAttribute('id'));
                });
                actions.push(`
            <p>I fixed the mobile menu so that it is compatible with the new publication (now multiple pages are generated instead of 1 single page for the whole website).</p>
          `);
            }
            resolve(actions);
        });
    }
    to2_2_9(version, doc) {
        return new Promise((resolve, reject) => {
            const actions = [];
            if (this.hasToUpdate(version, [2, 2, 9])) {
                const menuButton = doc.querySelector('.menu-button');
                if (menuButton) {
                    menuButton.parentElement.removeChild(menuButton);
                    actions.push(`
              <p>I removed the mobile menu as there is now a component for that.</p>
              <p><a target="_blank" href="https://github.com/silexlabs/Silex/wiki/Hamburger-menu">Read more about the Hamburger Menu component here</a>.</p>
            `);
                }
            }
            resolve(actions);
        });
    }
    to2_2_10(version, doc) {
        return new Promise((resolve, reject) => {
            const actions = [];
            if (this.hasToUpdate(version, [2, 2, 10])) {
                doc.body.classList.add(Constants_1.Constants.PREVENT_DRAGGABLE_CLASS_NAME, Constants_1.Constants.PREVENT_RESIZABLE_CLASS_NAME, Constants_1.Constants.PREVENT_SELECTABLE_CLASS_NAME);
                const changedSections = Array.from(doc.querySelectorAll(`.${Constants_1.Constants.TYPE_SECTION}`));
                changedSections.forEach((el) => el.classList.add(Constants_1.Constants.PREVENT_DRAGGABLE_CLASS_NAME, Constants_1.Constants.PREVENT_RESIZABLE_CLASS_NAME));
                const changedSectionsContent = Array.from(doc.querySelectorAll(`.${Constants_1.Constants.TYPE_SECTION}, .${Constants_1.Constants.TYPE_SECTION} .${Constants_1.Constants.TYPE_CONTAINER_CONTENT}`));
                changedSectionsContent.forEach((el) => el.classList.add(Constants_1.Constants.PREVENT_DRAGGABLE_CLASS_NAME));
                actions.push(`Changed the body and ${changedSections.length} sections with new CSS classes to <a href="https://github.com/silexlabs/stage/" target="_blank">the new stage component.</a>`);
                const changedElements = Array.from(doc.querySelectorAll(`[${Constants_1.Constants.TYPE_ATTR}]`));
                changedElements.forEach((el) => el.setAttribute(Constants_1.Constants.TYPE_ATTR, el.getAttribute(Constants_1.Constants.TYPE_ATTR) + '-element'));
                actions.push(`Updated ${changedElements.length} elements, changed their types to match the new version of Silex.`);
            }
            resolve(actions);
        });
    }
}
exports.default = BackwardCompat;
