"use strict";
Object.defineProperty(exports, "__esModule", { value: true });
const Constants_js_1 = require("../../Constants.js");
class DomTools {
    static transformPaths(dom, fn) {
        ['src', 'href'].forEach((attr) => {
            const elements = dom.window.document.querySelectorAll(`[${attr}]`);
            for (const el of elements) {
                if (el.tagName.toLowerCase() === 'a' || el.getAttribute('data-silex-href')) {
                    continue;
                }
                if (el.tagName.toLowerCase() === 'link' &&
                    el.getAttribute('rel').toLowerCase() !== 'stylesheet' &&
                    el.getAttribute('rel').toLowerCase() !== 'shortcut icon') {
                    continue;
                }
                if (el.hasAttribute('data-silex-static')) {
                    continue;
                }
                const val = el.getAttribute(attr);
                const newVal = fn(val, el, el.parentElement === dom.window.document.head);
                if (newVal) {
                    el.setAttribute(attr, newVal);
                }
            }
        });
        const tags = dom.window.document.querySelectorAll('style');
        const stylesheets = dom.window.document.styleSheets;
        const matches = [];
        for (let stylesheetIdx = 0; stylesheetIdx < stylesheets.length; stylesheetIdx++) {
            const stylesheet = stylesheets[stylesheetIdx];
            if (tags[stylesheetIdx]) {
                const tag = tags[stylesheetIdx];
                const cssText = DomTools.transformStylesheet(stylesheet, tag.parentElement === dom.window.document.head, fn);
                matches.push({
                    tag,
                    innerHTML: cssText,
                });
            }
        }
        matches.forEach(({ tag, innerHTML }) => tag.innerHTML = innerHTML);
        DomTools.transformJson(dom, fn);
    }
    static transformValueUrlKeyword(value, stylesheet, isInHead, fn) {
        if (typeof value === 'string' && value.indexOf('url(') === 0) {
            return `url(${value.replace(/url\('(.*)'\)|url\("(.*)"\)|url\((.*)\)/, (str, match1, match2, match3) => {
                const match = match1 || match2 || match3;
                return fn(match, stylesheet, isInHead) || match;
            })})`;
        }
        return null;
    }
    static transformStylesheet(stylesheet, isInHead, fn, isMediaQuerySubRule = false) {
        let cssText = '';
        for (const sheetOrRule of stylesheet.cssRules) {
            const rule = sheetOrRule;
            const sheet = sheetOrRule;
            if (rule.style) {
                for (const valName of Array.from(rule.style)) {
                    const value = rule.style[valName];
                    rule.style[valName] = DomTools.transformValueUrlKeyword(value, stylesheet, isInHead, fn) || value;
                }
            }
            else if (sheet.cssRules) {
                DomTools.transformStylesheet(sheet, isInHead, fn, true);
            }
            else {
            }
            if (!isMediaQuerySubRule) {
                cssText += rule.cssText;
            }
        }
        return cssText;
    }
    static transformJson(dom, fn) {
        const jsonData = DomTools.getProperties(dom.window.document);
        if (jsonData) {
            for (const dataObjName in jsonData) {
                const dataObj = jsonData[dataObjName];
                for (const elementId in dataObj) {
                    const elementData = dataObj[elementId];
                    for (const propName in elementData) {
                        const propValue = elementData[propName];
                        const valueUrlKeyword = DomTools.transformValueUrlKeyword(propValue, null, true, fn);
                        if (valueUrlKeyword) {
                            elementData[propName] = valueUrlKeyword;
                        }
                        else {
                            if (['src', 'href'].indexOf(propName) >= 0) {
                                elementData[propName] = fn(propValue) || propValue;
                            }
                        }
                    }
                }
            }
            DomTools.setProperties(dom.window.document, jsonData);
        }
    }
    static getProperties(doc) {
        const styleTag = doc.querySelector('.' + Constants_js_1.Constants.JSON_STYLE_TAG_CLASS_NAME);
        if (styleTag != null) {
            return JSON.parse(styleTag.innerHTML)[0];
        }
        return null;
    }
    static setProperties(doc, value) {
        const styleTag = doc.querySelector('.' + Constants_js_1.Constants.JSON_STYLE_TAG_CLASS_NAME);
        if (styleTag != null) {
            styleTag.innerHTML = JSON.stringify([value]);
        }
        else {
            console.error('Error: no JSON styles array found in the dom');
        }
    }
    static extractUserHeadTag(headString) {
        const regExp = new RegExp(Constants_js_1.Constants.HEAD_TAG_START + '([\\\s\\\S.]*)' + Constants_js_1.Constants.HEAD_TAG_STOP);
        const found = headString.match(regExp);
        if (found) {
            return {
                userHead: found[1],
                html: headString.replace(regExp, ''),
            };
        }
        return {
            userHead: '',
            html: headString,
        };
    }
    static insertUserHeadTag(htmlString, userHead) {
        if (userHead) {
            return htmlString.replace(/<\/head>/i, Constants_js_1.Constants.HEAD_TAG_START + userHead + Constants_js_1.Constants.HEAD_TAG_STOP + '</head>');
        }
        else {
            return htmlString;
        }
    }
}
exports.default = DomTools;
